library easy_encryption;

import 'package:cryptography/cryptography.dart';
import 'package:encrypt/encrypt.dart';

class EasyEncryption {
  final Map<String, int> _alphabetsIndex = {
    'a': 0,
    'b': 1,
    'c': 2,
    'd': 3,
    'e': 4,
    'f': 5,
    'g': 6,
    'h': 7,
    'i': 8,
    'j': 9,
    'k': 10,
    'l': 11,
    'm': 12,
    'n': 13,
    'o': 14,
    'p': 15,
    'q': 16,
    'r': 17,
    's': 18,
    't': 19,
    'u': 20,
    'v': 21,
    'w': 22,
    'x': 23,
    'y': 24,
    'z': 25,
    'A': 27,
    'B': 28,
    'C': 29,
    'D': 30,
    'E': 31,
    'F': 32,
    'G': 33,
    'H': 34,
    'I': 35,
    'J': 36,
    'K': 37,
    'L': 38,
    'M': 39,
    'N': 40,
    'O': 41,
    'P': 42,
    'Q': 43,
    'R': 44,
    'S': 45,
    'T': 46,
    'U': 47,
    'V': 48,
    'W': 49,
    'X': 50,
    'Y': 51,
    'Z': 52,
    ' ': 26,
    '!': 53,
    '#': 54,
    '\$': 55,
    '%': 56,
    '&': 57,
    '@': 58,
    '(': 59,
    ')': 60,
    '-': 61,
    '_': 62,
    '+': 63,
    '/': 64,
    '*': 65,
    ':': 66,
    ';': 67,
    '?': 68,
    '.': 69,
    ',': 70,
    '|': 71,
    '^': 72,
    '[': 73,
    ']': 74,
    '{': 75,
    '}': 76,
    '<': 77,
    '>': 78,
    '=': 79,
    '~': 80,
    '0': 81,
    '1': 82,
    '2': 83,
    '3': 84,
    '4': 85,
    '5': 86,
    '6': 87,
    '7': 88,
    '8': 89,
    '9': 90,
  };

  final Map<int, String> _sBox = {
    4: 'a',
    36: 'b',
    50: 'c',
    48: 'd',
    62: 'e',
    46: 'f',
    55: 'g',
    30: 'h',
    54: 'i',
    5: 'j',
    38: 'k',
    12: 'l',
    34: 'm',
    53: 'n',
    72: 'o',
    13: 'p',
    67: 'q',
    40: 'r',
    20: 's',
    52: 't',
    51: 'u',
    14: 'v',
    77: 'w',
    27: 'x',
    6: 'y',
    79: 'z',
    16: 'A',
    8: 'B',
    17: 'C',
    42: 'D',
    49: 'E',
    19: 'F',
    78: 'G',
    61: 'H',
    76: 'I',
    56: 'J',
    44: 'K',
    15: 'L',
    43: 'M',
    69: 'N',
    7: 'O',
    41: 'P',
    33: 'Q',
    71: 'R',
    74: 'S',
    73: 'T',
    28: 'U',
    37: 'V',
    10: 'W',
    21: 'X',
    18: 'Y',
    70: 'Z',
    75: ' ',
    80: '!',
    32: '#',
    26: '\$',
    65: '%',
    11: '&',
    25: '@',
    23: '(',
    64: ')',
    1: '-',
    68: '_',
    45: '+',
    66: '/',
    35: '*',
    24: ':',
    3: ';',
    39: '?',
    57: '.',
    9: ',',
    58: '|',
    63: '^',
    2: '[',
    60: ']',
    31: '{',
    59: '}',
    22: '<',
    29: '>',
    47: '=',
    0: '~',
    89: '0',
    84: '1',
    85: '2',
    81: '3',
    88: '4',
    82: '5',
    87: '6',
    90: '7',
    86: '8',
    83: '9',
  };

  final _iv = IV.fromLength(16);

  Future<String> encryptData({required String data}) async {
    // Hashing Message
    String _hashedMessage = await _hashMessage(data);

    // Generating Key
    String _generatedKey = _generateKey(data);

    final _key = Key.fromUtf8(_generatedKey);

    // Encrypting Message
    final _encrypter = Encrypter(AES(_key));

    final _encrypted = _encrypter.encrypt(_hashedMessage, iv: _iv);

    // Embedding key in cipher text and sending it as final
    final _cipherText = _embeddKey(_encrypted.base64, _key.base64);

    // print("Cipher Text: " + cipherText);
    return _cipherText;
  }

  Future<String> _hashMessage(String message) async {
    final sink = Sha512().newHashSink();
    final messageInput = _getAlphabetArray(message);
    sink.add(messageInput);
    sink.close();
    final hash = await sink.hash();

    String hashedMessage = message;

    // Hash message is seperated with | in plain text message
    for (int i = 0; i < hash.bytes.length; i++) {
      hashedMessage += "|" + hash.bytes[i].toString();
    }

    return hashedMessage;
  }

  List<int> _getAlphabetArray(String string) {
    List<int> result = [];
    for (int i = 0; i < string.length; i++) {
      result.add(_alphabetsIndex[string[i]]!);
    }
    return result;
  }

  String _generateKey(String message) {
    String key = "";
    String length32BitLength = "";

    if (message.length > 32) {
      length32BitLength = message.substring(1, 33);
    } else {
      if (message.length != 32) {
        String padding = "";
        for (int i = message.length; i < 32; i++) {
          padding +=
              _alphabetsIndex.keys.elementAt((i * 2) % _alphabetsIndex.length);
        }
        length32BitLength += message + padding;
      } else {
        length32BitLength = message;
      }
    }

    for (int i = 0; i < length32BitLength.length / 2; i++) {
      String newString = length32BitLength[length32BitLength.length - 1];
      newString += length32BitLength.substring(0, 31);
      length32BitLength = newString;
    }

    for (int i = 0; i < length32BitLength.length; i++) {
      int index = _alphabetsIndex[length32BitLength[i]]!;
      key += _sBox[index]!;
    }

    return key;
  }

  String _embeddKey(String cipherText, String key) {

    String newStr = cipherText.substring(0, (cipherText.length / 2).round()) +
        key +
        cipherText.substring(
            (cipherText.length / 2).round(), cipherText.length);

    return newStr;
  }

  Future<String> decryptData({required String data}) async {
    // Extract Embedded key and remove it and get cipher text
    final _result = _extractEmbeddedKeyAndCipherText(data);

    final _key = _result["key"];
    final _extractedCipherText = _result["cipherText"];
    // Decrypt the cipher text

    final _plainTextWithHash = _decryptCipherText(_extractedCipherText!, _key!);
    // Check if the hash is same or else message is tampered

    final _isHashValid = await _checkHashValid(_plainTextWithHash);

    // return the decrypted plain text

    String _plainText = "";

    if (_isHashValid) {
      _plainText = _extractPlainText(_plainTextWithHash);
    }

    return _plainText;
  }

  Map<String, String> _extractEmbeddedKeyAndCipherText(String cipherText) {
    int startIndex = (((cipherText.length - 44) / 2)).round();
    String key = "";
    String sampleCipherText = cipherText.substring(0, startIndex) +
        cipherText.substring(startIndex + 44, cipherText.length);

    key = cipherText.substring(startIndex, startIndex + 44);

    return {"cipherText": sampleCipherText, "key": key};
  }

  String _decryptCipherText(String cipherText, String key) {
    final extractedKey = Key.fromBase64(key);
    final encrypter = Encrypter(AES(extractedKey));
    final encrypted = Encrypted.fromBase64(cipherText);
    final plainText = encrypter.decrypt(encrypted, iv: _iv);
    return plainText;
  }

  Future<bool> _checkHashValid(String plainText) async {
    final sink = Sha512().newHashSink();

    String message = "";
    String actualHash = "";

    for (int i = 0; i < plainText.length; i++) {
      if (plainText[i] == "|") {
        message = plainText.substring(0, i);
        actualHash = plainText.substring(i, plainText.length);
        break;
      }
    }
    final messageInput = _getAlphabetArray(message);
    sink.add(messageInput);
    sink.close();
    final hash = await sink.hash();

    String calculatedHash = "";

    for (int i = 0; i < hash.bytes.length; i++) {
      calculatedHash += "|" + hash.bytes[i].toString();
    }

    return actualHash == calculatedHash;
  }

  String _extractPlainText(String plainTextWithHash) {
    String plainText = "";
    for (int i = 0; i < plainTextWithHash.length; i++) {
      if (plainTextWithHash[i] == "|") {
        plainText = plainTextWithHash.substring(0, i);
        break;
      }
    }
    return plainText;
  }
}
